/****************************************************************************
 *  
 * Project: Olimex AVR ISP with STK500 version 2 protocol support.
 *
 * Copyright by Olimex Ltd. All rights reserved.
 * 
 * File: dataflash.c
 * Description: ATMEL Serial Dataflash access routines.
 * Developer: Dimitar Dimitrov ( dinuxbg@gmail.com )
 *
 * Last change: $Date: 2007-12-11 17:38:11 +0200 (вторник, 11 Декември 2007) $
 * Revision: $Revision: 86 $
 * Id: $Id: dataflash.c 86 2007-12-11 15:38:11Z dimitar $
 *
 ****************************************************************************/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdint.h>

#include "dataflash.h"
#include "swspi.h"

/* Dataflash read commands .*/
#define DF_CMD_MAINMEM_PAGE_READ	0xD2
#define DF_CMD_CONTARRAY_READ_OLDV	0xE8
#define DF_CMD_CONTARRY_READ_LF		0x03
#define DF_CMD_CONTARRY_READ_HF		0x0B
#define DF_CMD_BUF1_READ_LF		0xD1
#define DF_CMD_BUF2_READ_LF		0xD3
#define DF_CMD_BUF1_READ_HF		0xD4
#define DF_CMD_BUF2_READ_HF		0xD6

/* Dataflash program and erase commands .*/
#define DF_CMD_BUF1_WRITE		0x84
#define DF_CMD_BUF2_WRITE		0x87
#define DF_CMD_BUF1_TO_MEMPG_WI_ERASE	0x83
#define DF_CMD_BUF2_TO_MEMPG_WI_ERASE	0x86
#define DF_CMD_BUF1_TO_MEMPG_WO_ERASE	0x88
#define DF_CMD_BUF2_TO_MEMPG_WO_ERASE	0x89
#define DF_CMD_PAGE_ERASE		0x81
#define DF_CMD_BLOCK_ERASE		0x50
#define DF_CMD_SECTOR_ERASE		0x7C
#define DF_CMD_CHIP_ERASE_PART1		0xC7
#define DF_CMD_CHIP_ERASE_PART2		0x94
#define DF_CMD_CHIP_ERASE_PART3		0x80
#define DF_CMD_CHIP_ERASE_PART4		0x9A
#define DF_CMD_MEMPG_PROG_THROUGH_BUF1	0x82
#define DF_CMD_MEMPG_PROG_THROUGH_BUF2	0x85

/* Dataflash Protection and Security Commands - NOT USED*/

/* Dataflash Additional Commands */
#define DF_CMD_MEMPG_TO_BUF1_TRANSFER	0x53
#define DF_CMD_MEMPG_TO_BUF2_TRANSFER	0x55
#define DF_CMD_MEMPG_TO_BUF1_COMPARE	0x60
#define DF_CMD_MEMPG_TO_BUF2_COMPARE	0x61
#define DF_CMD_AUTO_PAGE_REWRITE_BUF1	0x58
#define DF_CMD_AUTO_PAGE_REWRITE_BUF2	0x59
#define DF_CMD_DEEP_POWER_DOWN		0xB9
#define DF_CMD_RESUME_FROM_POWER_DOWN	0xAB
#define DF_CMD_STATUS_REGISTER_READ	0xD7
#define DF_CMD_READ_MANIF_DEV_ID	0x9F

/** Dataflash page address/offset number of bits. */
#define DF_PAGE_ADDR_NBITS		9
#define DF_PAGE_SIZE			(1<<DF_PAGE_ADDR_NBITS)
#define DF_PAGE_ADDR_MASK		(DF_PAGE_SIZE-1)
#define DF_BLOCK_SIZE			(8*DF_PAGE_SIZE)
#define DF_NUM_BLOCKS			(DATAFLASH_SIZE/DF_BLOCK_SIZE)

static uint8_t dataflash_manifdevid[4] PROGMEM = {
	0x1f,	/* Manifacturer ID: Atmel */
	0x26,	/* Device id: Dataflash, 16-Mbit */
	0x00,	/* Device id2: 1-bit Cell Technology, initial version. */
	0x00	/* No extended information */
};

/**
 * @brief Wait for ready using polling of the dataflash status register.
 */
static void wait_busy(void)
{
	uint8_t b;

	SWSPI_CS(0);
	swspi_xdata(DF_CMD_STATUS_REGISTER_READ);
	do {
		b = swspi_xdata(0);
	} while(~b & (1<<7));
	SWSPI_CS(1);
}

/**
 * @brief Get dataflash status register contents.
 * @return Dataflash status register value.
 */
static uint8_t read_status_reg(void)
{
	uint8_t b;

	SWSPI_CS(0);
	swspi_xdata(DF_CMD_STATUS_REGISTER_READ);
	b = swspi_xdata(0);
	SWSPI_CS(1);
	return b;
}

/**
 * @brief Program the "Power of 2" binary page size option.
 * @warning BE VERY CAREFUL WITH THAT COMMAND! ONCE PROGRAMMED THIS OPTION CAN
 * NOT BE RESTORED BACK!
 */
static void dataflash_program_powerof2_option(void)
{
	SWSPI_WP(1);
	SWSPI_CS(0);
	swspi_xdata(0x3D);
	swspi_xdata(0x2A);
	swspi_xdata(0x80);
	swspi_xdata(0xA6);
	SWSPI_CS(1);

	wait_busy();
	SWSPI_WP(0);
}

/** 
 * @brief Initialize the connected dataflash.
 * @return Zero for success, negative value on error.
 */
int8_t dataflash_init(void)
{
	int status = 0;
	uint8_t i, b;

	swspi_init();

	SWSPI_CS(0);
	swspi_xdata(DF_CMD_READ_MANIF_DEV_ID);
	for(i=0; i<sizeof(dataflash_manifdevid); ++i) {
		b = pgm_read_byte(dataflash_manifdevid+i);
		if(b != swspi_xdata(0x00)) status = -1;
	}
	SWSPI_CS(1);

	if(~read_status_reg() & 1) {
		/* Dataflash uses 528-byte page size. Reconfigure it for 
		 * a more "sane" 512-byte one. */
		dataflash_program_powerof2_option();
		status = -2;
	}

	return status;
}

/** 
 * Erase the DataFlash chip. 
 *
 * @note We cannot use the chip erase command because of chip errata!
 */
void dataflash_chip_erase(void)
{
	uint32_t i;

	SWSPI_WP(1);
	for(i=0; i<DF_NUM_BLOCKS; i += DF_BLOCK_SIZE) {
		SWSPI_CS(0);
		swspi_xdata(DF_CMD_BLOCK_ERASE);
		swspi_xdata((i>>16) & 0xff);
		swspi_xdata((i>>8) & 0xff);
		swspi_xdata((i>>0) & 0xff);
		SWSPI_CS(1);
		wait_busy();
	}
	SWSPI_WP(0);
}



/** Program buffer 1 to the current main memory page. Use auto-erase dataflash
 * feature. 
 *
 * @note It's user's responsibility to wait for write-complete!
 * */
static void program_buf1_to_mainmempg(uint32_t addr)
{
	SWSPI_CS(0);
	swspi_xdata(DF_CMD_BUF1_TO_MEMPG_WI_ERASE);
	swspi_xdata((addr>>16) & 0xff);
	swspi_xdata((addr>>8) & 0xff);
	swspi_xdata((addr>>0) & 0xff);
	SWSPI_CS(1);
}


/** Read main memory page at @addr and store it in dataflash's buffer 1. */
static void transfer_mainmempg_to_buf1(uint32_t addr)
{
	SWSPI_CS(0);
	swspi_xdata(DF_CMD_MEMPG_TO_BUF1_TRANSFER);
	swspi_xdata((addr>>16) & 0xff);
	swspi_xdata((addr>>8) & 0xff);
	swspi_xdata((addr>>0) & 0xff);
	SWSPI_CS(1);
	wait_busy();
}

/** 
 * Position the write pointer to buffer 1 at offset @addr. After that all
 * clocked data bytes will be written into subsequent cells of buffer 1.
 *
 * @note After function returns the dataflash is in active mode, expecting
 * bytes to be written to its buffer. It's user's responsibility to end the
 * write by pulling CS high.
 * */
static void position_buf1_write(uint16_t offset)
{
	SWSPI_CS(0);
	swspi_xdata(DF_CMD_BUF1_WRITE);
	swspi_xdata(0x00);		/* don't care */
	swspi_xdata((offset>>8) & 0xff);
	swspi_xdata((offset>>0) & 0xff);
}



#if DATAFLASH_STREAM_API

/**
 * @brief Read @a nbytes from serial dataflash and store them in @a buf.
 * @param buf Buffer where to store the read bytes.
 * @param nbytes Number of bytes to read.
 * @param addr Dataflash read start address.
 */
void dataflash_read(uint8_t *buf, uint16_t nbytes, uint32_t addr)
{
	SWSPI_CS(0);
	swspi_xdata(DF_CMD_CONTARRY_READ_HF);
	swspi_xdata((addr>>16) & 0xff);
	swspi_xdata((addr>>8) & 0xff);
	swspi_xdata((addr>>0) & 0xff);
	swspi_xdata(0);	/* dummy byte */
	while(nbytes) {
		--nbytes;
		*buf++ = swspi_xdata(0x00);
	}
	SWSPI_CS(1);
}

/**
 * @brief Read up to @a nbytes from serial dataflash and compare them against
 * the contents of @a buf.
 * @param buf Buffer which to compare against.
 * @param nbytes Maximum number of bytes to compare.
 * @param addr Dataflash read start address.
 * @return Zero if contents of dataflash and input buffer match, non-zero 
 * otherwise.
 */
int8_t dataflash_compare(const uint8_t *buf, uint16_t nbytes, uint32_t addr)
{
	int8_t stat = 0;

	SWSPI_CS(0);
	swspi_xdata(DF_CMD_CONTARRY_READ_HF);
	swspi_xdata((addr>>16) & 0xff);
	swspi_xdata((addr>>8) & 0xff);
	swspi_xdata((addr>>0) & 0xff);
	swspi_xdata(0);	/* dummy byte */
	while(nbytes && !stat) {
		--nbytes;
		if(*buf++ != swspi_xdata(0x00)) {
			stat = -1;
		}
	}
	SWSPI_CS(1);
	return stat;
}

/**
 * Initialize a continuous read operation, starting with address @a addr . 
 * This process must be terminated by calling @a dataflash_contread_stop .
 *
 * @param addr Starting read address.
 */
void dataflash_contread_start(const uint32_t addr)
{
	SWSPI_CS(0);
	swspi_xdata(DF_CMD_CONTARRY_READ_HF);
	swspi_xdata((addr>>16) & 0xff);
	swspi_xdata((addr>>8) & 0xff);
	swspi_xdata((addr>>0) & 0xff);
	swspi_xdata(0);	/* dummy byte */
}

/**
 * Fetch the next byte while in a continuous read mode operation.
 * 
 * @return The read byte.
 */
uint8_t dataflash_contread_fetch(void)
{
	return swspi_xdata(0x00);
}

/** Stop the continous read operation. */
void dataflash_contread_stop(void)
{
	SWSPI_CS(1);
}



/** Dataflash buffer state. */
static enum {
	ST_IDLE,	/**< Idle and ready for shutdown (initial state) */
	ST_BUFEDIT	/**< Page has been loaded in buffer and is being 
			  modified */
} df_state;

/** 
 * Address of the page that is currently loaded in buffer 1. Valid only when
 * df_state equals ST_BUFEDIT
 */
static uint32_t buf1_pg_addr;

/** Flush the current open page from buffer 1 to the main flash memory. */
void dataflash_flush(void)
{
	if(df_state == ST_BUFEDIT) {
		SWSPI_WP(1);
		program_buf1_to_mainmempg(buf1_pg_addr);
		wait_busy();
		df_state = ST_IDLE;
		SWSPI_WP(0);
	}
}


/**
 * @brief Write @a nbytes from @buf to serial dataflash.
 *
 * @param buf Buffer where to read bytes.
 * @param nbytes Number of bytes to write.
 * @param addr Dataflash write start address.
 */
void dataflash_write(const uint8_t *buf, uint16_t nbytes, uint32_t addr)
{
	uint16_t pg_offset;

	SWSPI_WP(1);
	pg_offset = addr & DF_PAGE_ADDR_MASK;

	/* If user tries to write to a different page then flush the current
	 * one and update our internal state .*/
	if((addr & ~DF_PAGE_ADDR_MASK) != (buf1_pg_addr & ~DF_PAGE_ADDR_MASK)) {
		if(df_state == ST_BUFEDIT) dataflash_flush();
		buf1_pg_addr = addr;
		df_state = ST_IDLE;
	}

	/* Enter buffer write mode if neccessary */
	if(df_state == ST_BUFEDIT) {
		position_buf1_write(addr);
	}

	while(nbytes) {
		switch(df_state) {
			case ST_IDLE:
				/* Load page into buffer 1 */
				transfer_mainmempg_to_buf1(addr);
				position_buf1_write(addr);
				buf1_pg_addr = addr;
				df_state = ST_BUFEDIT;
				pg_offset = addr & DF_PAGE_ADDR_MASK;
				break;
			case ST_BUFEDIT:
				/* send byte to buffer 1 */
				swspi_xdata(*buf++);
				--nbytes;
				/* increment page offset and take actions if
				 * we're at the end of the page */
				if(pg_offset == DF_PAGE_ADDR_MASK) {
					pg_offset = 0;
					df_state = ST_IDLE;
					SWSPI_CS(1);
					/* write back page */
					program_buf1_to_mainmempg(addr);
					addr = (addr | DF_PAGE_ADDR_MASK) + 1;
					wait_busy();
				} else {
					++pg_offset;
				}
				break;
			default:
				df_state = ST_IDLE;
		}

	}
	SWSPI_CS(1);
	SWSPI_WP(0);
}


#endif /* DATAFLASH_STREAM_API */




/*------------------------------- PAGE API ---------------------------------*/

#if DATAFLASH_PAGE_API

/** 
 * Position the read pointer to buffer 1 at offset @addr. After that all
 * clocked data bytes will be written into subsequent cells of buffer 1.
 *
 * @note After function returns the dataflash is in active mode, expecting
 * bytes to be read from its buffer. It's user's responsibility to end the
 * operation by pulling CS high.
 * */
static void position_buf1_read(uint16_t offset)
{
	SWSPI_CS(0);
	swspi_xdata(DF_CMD_BUF1_READ_LF);
	swspi_xdata(0x00);		/* don't care */
	swspi_xdata((offset>>8) & 0xff);
	swspi_xdata((offset>>0) & 0xff);
}



/** Read page @a pagenum and store it in Dataflash buffer 1. */
void dataflash_mem2buf1(uint32_t pagenum)
{
	transfer_mainmempg_to_buf1(pagenum * DF_PAGE_SIZE);
}


/** 
 * Read @a nbytes number of bytes from Dataflash buffer 1 into @a dstbuf,
 * starting from offset @a offset.
 */
void dataflash_rbuf1(uint8_t *dstbuf, uint16_t nbytes, uint16_t offset)
{
	position_buf1_read(offset);
	while(nbytes) {
		--nbytes;
		*dstbuf++ = swspi_xdata(0x00);
	}
	SWSPI_CS(1);
}


/**
 * Write @a nbytes from @a srcbuf to Dataflash buffer 1, starting from offset
 * @a offset.
 */
void dataflash_wbuf1(const uint8_t *srcbuf, uint16_t nbytes, uint16_t offset)
{
	position_buf1_write(offset);
	while(nbytes) {
		--nbytes;
		swspi_xdata(*srcbuf++);
	}
	SWSPI_CS(1);
}

/** Store Dataflash buffer 1 to page @a pagenum */
void dataflash_buf12mem(uint32_t pagenum)
{
	SWSPI_WP(1);
	program_buf1_to_mainmempg(pagenum * DF_PAGE_SIZE);
	wait_busy();
	SWSPI_WP(0);
}


#endif /* DATAFLASH_PAGE_API */

