/****************************************************************************
 *  
 * Project: Olimex AVR ISP with MKII protocol support.
 *
 * Copyright by Olimex Ltd. All rights reserved.
 * 
 * File: usb_hw.h
 * Description: Hardware-dependant USB functionality.
 * Developer: Dimitar Dimitrov ( dinuxbg __on__ gmail.com )
 *
 * Last change: $Date: 2007-12-19 17:50:39 +0200 (сряда, 19 Декември 2007) $
 * Revision: $Revision: 95 $
 * Id: $Id: usb_hw.h 95 2007-12-19 15:50:39Z dimitar $
 *
 ****************************************************************************/

#ifndef USB_HW_H
#define USB_HW_H

/*-------------------------------- DEVICE --------------------------------*/

/** Enable USB 3.3V regulator. */
#define usb_enable_regulator()	(REGCR &= ~(1<<REGDIS))

/** Enable USB controller .*/
#define usb_macro_enable()	( USBCON = (1<<USBE))

/** Disable USB controller. */
#define usb_macro_disable()	( USBCON = (0<<USBE))

/** Freeze USB clock. */
#define usb_freeze_clock()	( USBCON |= (1<<FRZCLK))

/** Unfreeze the USB clock. */
#define usb_resume_clock()	( USBCON &= ~(1<<FRZCLK))

/** Select the right USB PLL configuration and start it. */
#if F_CPU == 8000000UL
  #define usb_pll_start()	( PLLCSR = (0<<PLLP1)|(0<<PLLP0)|(1<<PLLE))
#elif F_CPU == 16000000UL
  #define usb_pll_start()	( PLLCSR = (0<<PLLP1)|(1<<PLLP0)|(1<<PLLE))
#else
  #error USB PLL requires either 8MHz or 16MHz CPU clock!
#endif

/** Stop USB PLL. */
#define usb_pll_stop()		( PLLCSR = (0<<PLLE))

/** Check whether USB PLL has locked. */
#define usb_pll_locked()	(PLLCSR & (1<<PLOCK))

/** Attach USB device controller. */
#define usb_attach()		( UDCON &= ~(1<<DETACH))

/** Check whether USB device controller is detached. */
#define usb_detached()		( UDCON & (1<<DETACH) )

/** Detach USB device controller. */
#define usb_detach()		( UDCON |= (1<<DETACH))

/** Fetch the full 11bits from the USB Frame Counter register pair. */
#define usb_get_frcuntr()	( UDFNUM )

/** Fetch the LSB 8bits from the USB Frame Counter. */
#define usb_get_frcntr_8()	( UDFNUML )

/** Fetch General USB interrupt. */
#define usb_genintr_fetch()	(UDINT & UDIEN)

/** Clear General USB interrupt masked bits. */
#define usb_genintr_clear(MASK)	( UDINT = (uint8_t)(~(MASK)))

/** RESUME interrupt flag raised. */
#define usb_intst_resume(v)	( (v) & (1<<EORSMI) )
#define usb_inten_resume()	( UDIEN |= (1<<EORSME))
#define usb_intdis_resume()	( UDIEN &= ~(1<<EORSME))
#define usb_intclr_resume()	( UDINT &= ~(1<<EORSMI))

/** WAKE UP interrupt flag raised. */
#define usb_intst_wakeup(v)	( (v) & (1<<WAKEUPI) )
#define usb_inten_wakeup()	( UDIEN |= (1<<WAKEUPE))
#define usb_intdis_wakeup()	( UDIEN &= ~(1<<WAKEUPE))
#define usb_intclr_wakeup()	( UDINT &= ~(1<<WAKEUPI))

/** START-OF-FRAME interrupt flag raised. */
#define usb_intst_sof(v)	( (v) & (1<<SOFI) )
#define usb_inten_sof()		( UDIEN |= (1<<SOFE))
#define usb_intdis_sof()	( UDIEN &= ~(1<<SOFE))

/** END-OF-RESET interrupt flah raised. */
#define usb_intst_reset(v)	( (v) & (1<<EORSTI) )
#define usb_inten_reset()	( UDIEN |= (1<<EORSTE))
#define usb_intdis_reset()	( UDIEN &= ~(1<<EORSTE))

/** SUSPEND interrupt flah raised. */
#define usb_intst_suspend(v)	( (v) & (1<<SUSPI) )
#define usb_inten_suspend()	( UDIEN |= (1<<SUSPE))
#define usb_intdis_suspend()	( UDIEN &= ~(1<<SUSPE))
#define usb_intclr_suspend()	( UDINT &= ~(1<<SUSPI))

/** Load address but do not enable it. */
#define usb_devaddr_load(A)	( UDADDR = (UDADDR & (1<<ADDEN)) | (A))

/** Check whether device is addressed. */
#define usb_dev_addressed()	(UDADDR & (1<<ADDEN))

/** Enable the loaded address. */
#define usb_devaddr_latch()	( UDADDR |= (1<<ADDEN))

/*-------------------------------- ENDPOINT --------------------------------*/

/** Select endpoint for the following USB operations. */
#define usb_ep_select(epnum)	( UENUM = (uint8_t)(epnum))

/** Enable endpoint. */
#define usb_ep_enable()		( UECONX = (1<<EPEN))

/** Endpoint enabled. */
#define usb_ep_enabled() 	( UECONX & (1<<EPEN) )

/** Disable endpoint. */
#define usb_ep_disable()	( UECONX = (0<<EPEN))

/** Allocate memory for the selected endpoint. */
#define usb_epmem_allocate()	( UECFG1X |= (1<<ALLOC))

/** Deallocate memory for the selected endpoint. */
#define usb_epmem_deallocate()	( UECFG1X &= ~(1<<ALLOC))

#define usb_epsize_conv(S)	((S)==8?0x00 : ((S)==16?0x01 : ((S)==32?0x02 : ((S)==64? 0x03 : (S)))))

/** Builds configuration bytes and tries to activate the endpoint. */
#define usb_ep_activate(epnum,eptype,epdir,epsize,dblbank) 	\
	( usb_ep_select(epnum), \
	  usb_ep_enable(), \
	  usb_ep_activate_conf( \
		  (((eptype)<<EPTYPE0) | ((epdir) << EPDIR)), \
		  (((usb_epsize_conv(epsize))<<EPSIZE0) | ((dblbank)<<EPBK0) | (1<<ALLOC))))

/** Reset the currently selected endpoint . */
#define usb_ep_reset(epnum)	\
	do { \
		UERST |= (uint8_t)((uint8_t)1<<((uint8_t)(epnum)+EPRST0)); \
		UERST &= (uint8_t)~((uint8_t)1<<((uint8_t)(epnum)+EPRST0)); \
	} while(0)


/** Reset the currently selected endpoint datatoggle sequence. */
#define usb_ep_reset_datatoggle()	( UECONX |= (1<<RSTDT))

/** Check STALL for the selected endpoint. */
#define usb_ep_stall_enabled()	( UECONX & (1<<STALLRQ))

/** Enable STALL for the selected endpoint. */
#define usb_ep_stall_enable()	( UECONX |= (1<<STALLRQ))

/** Disable STALL for the selected endpoint. */
#define usb_ep_stall_disable()	( UECONX |= (1<<STALLRQC))

/*-------------------------------- TRANSFER --------------------------------*/

/** Bit definition somehow missed in the system headers. */
#define KILLBK			RXOUTI

#define usb_epfifo_bytecount()	( UEBCLX )
/** Check whether the FIFO is available for Read/Write. */
#define usb_epfifo_ready()	( UEINTX & (1<<RWAL))

/** Push a value on the endpoint's FIFO. */
#define usb_epfifo_push(A)	( UEDATX = (A))

/** Pop a value from the endpoint's FIFO. */
#define usb_epfifo_pop()	(UEDATX)

/** Check whether a SETUP packet has been received. */
#define usb_setupreq_ready()	(UEINTX & (1<<RXSTPI))

/** 
 * Send acknowledge to the received SETUP packet. As a SETUP packet must begin
 * with initialized state (empty fifos, no received/transmitted stuff, we clear
 * most of the status bits. The TXINI flag is left because clearing it would
 * send a ZLP on the next IN request.
 */
#define usb_setupreq_sendack()	( UEINTX = (1<<TXINI))
/*#define usb_setupreq_sendack()	( UEINTX = ~(1<<RXSTPI) & ~(1<<KILLBK))*/

/** Clear a NAK for an OUT packet. */
#define usb_ctrlout_nakclr()	( UEINTX = ~(1<<NAKOUTI) & ~(1<<KILLBK))

/** Check whether an OUT packet has been received and NAK has been sent. */
#define usb_ctrlout_naksent()	(UEINTX & (1<<NAKOUTI))

/** Check whether an OUT packet has been received. */
#define usb_ctrlout_ready()	(UEINTX & (1<<RXOUTI))

/** Send acknowledge to the received OUT packet. */
#define usb_ctrlout_sendack()	( UEINTX = ~(1<<RXOUTI))

/** Clear a NAK for an IN packet. */
#define usb_ctrlin_nakclr()	( UEINTX = ~(1<<NAKINI) & ~(1<<KILLBK))

/** Check whether an IN packet has been received and NAK has been sent. */
#define usb_ctrlin_naksent()	(UEINTX & (1<<NAKINI))

/** Send the contents of endpoint's FIFO contents with the next IN request. 
 *
 * @note What on Earth is the KILLBK bit doing at the same place as RXOUTI 
 * bit 2 in UEINTX? And why it is the only bit affected by software writing
 * one to it?
 */
#define usb_ctrlin_send()	( UEINTX = ~(1<<TXINI) & ~(1<<KILLBK))

/** Check whether Control IN is ready for new data. */
#define usb_ctrlin_ready()	(UEINTX & (1<<TXINI))

#define usb_out_ready()		(UEINTX & (1<<RXOUTI))
#define usb_out_sendack()	(UEINTX = ~(1<<RXOUTI))
#define usb_out_clrbank()	(UEINTX &= ~(1<<FIFOCON))
#define usb_in_ready()		(UEINTX & (1<<TXINI))
#define usb_in_sendack()	(UEINTX = ~(1<<TXINI) & ~(1<<KILLBK))
#define usb_in_clrbank()	(UEINTX = (uint8_t)~(1<<FIFOCON) & ~(1<<KILLBK))
#define usb_in_naksent()	(UEINTX & (1<<NAKINI))
#define usb_in_nakclr()		( UEINTX = ~(1<<NAKINI) & ~(1<<KILLBK))

/*------------------------------- FUNCTIONS -------------------------------*/

/**
 * Do an atomic operation @a OP . This will disable interrupts, do @a OP and 
 * then restore the previus interrupt setting.
 */
#define USB_ATOMIC_OPER(OP)	\
	do { \
		uint8_t sreg = SREG; \
		cli(); \
		(OP); \
		SREG = sreg; \
	} while(0)

/**
 * Send data, located in Program Flash, using Control IN transfers.
 * This helper function is used by the USB core module but could be useful
 * to class-specific code, too.
 *
 * @param stppkt Pointer to the just received SETUP packet.
 * @param data Data that should be transmitted (should be in ROM)
 * @param nbytes Number of data bytes in @a data .
 * @param fetch_data Function that fetches one byte from @a data .
 *
 * @note This is made as an ugly macros because it might be useful not only 
 * for core USB but also for class-specific USB modules. Making it a plaing
 * function would bloat code size for avr-isp500-mass. 
 */
#define usb_ctrlin_send_data(stppkt, data, nbytes, fetch_data)	\
do { \
	uint8_t need_zlp; \
\
	if(nbytes >= stppkt->length) { \
		nbytes = stppkt->length; \
		need_zlp = 0; \
	} else { \
		if((nbytes & (USB_CTRLEP_MAX_PACKET_SIZE-1)) == 0) { \
			need_zlp = 1; \
		} else { \
			need_zlp = 0; \
		} \
	} \
\
	/* now send data */ \
\
	while(nbytes && !usb_ctrlout_naksent()) { \
		uint8_t i; \
\
		for(	i=0; \
			i<USB_CTRLEP_MAX_PACKET_SIZE && nbytes; \
			++i, --nbytes)  \
		{ \
			usb_epfifo_push(fetch_data(data)); \
			++data; \
		} \
\
		if(!usb_ctrlout_naksent()) { \
			usb_ctrlin_send(); \
			while(!usb_ctrlin_ready() && !usb_ctrlout_naksent()); \
		} \
	} \
\
	if(need_zlp && !usb_ctrlout_naksent()) { \
		usb_ctrlin_send(); \
		while(!usb_ctrlin_ready()); \
	} \
\
	while(!(usb_ctrlout_naksent())); \
	usb_ctrlout_nakclr(); \
	usb_ctrlout_sendack(); \
} while(0)

extern int8_t usb_ep_activate_conf(uint8_t config0x, uint8_t config1x);

#endif /* USB_HW_H */
